"""NotebookNode - adding attribute access to dicts"""

from ._struct import Struct

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


class NotebookNode(Struct):
    """A dict-like node with attribute-access"""

    def __setitem__(self, key, value):
        if isinstance(value, Mapping) and not isinstance(value, NotebookNode):
            value = from_dict(value)
        super().__setitem__(key, value)

    def update(self, *args, **kwargs):
        """
        A dict-like update method based on CPython's MutableMapping `update`
        method.
        """
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))
        if args:
            other = args[0]
            if isinstance(other, Mapping):
                for key in other:
                    self[key] = other[key]
            elif hasattr(other, "keys"):
                for key in other.keys():
                    self[key] = other[key]
            else:
                for key, value in other:
                    self[key] = value
        for key, value in kwargs.items():
            self[key] = value


def from_dict(d):
    """Convert dict to dict-like NotebookNode

    Recursively converts any dict in the container to a NotebookNode.
    This does not check that the contents of the dictionary make a valid
    notebook or part of a notebook.
    """
    if isinstance(d, dict):
        return NotebookNode({k: from_dict(v) for k, v in d.items()})
    elif isinstance(d, (tuple, list)):
        return [from_dict(i) for i in d]
    else:
        return d



from unittest import TestCase

from nbformat.v1.nbjson import reads, writes

from .nbexamples import nb0


class TestJSON(TestCase):
    def test_roundtrip(self):
        s = writes(nb0)
        self.assertEqual(reads(s), nb0)



"""Tests for the Python API for composing notebook elements"""

from nbformat.v4.nbbase import (
    NotebookNode,
    nbformat,
    new_code_cell,
    new_markdown_cell,
    new_notebook,
    new_output,
    new_raw_cell,
)


def test_empty_notebook():
    nb = new_notebook()
    assert nb.cells == []
    assert nb.metadata == NotebookNode()
    assert nb.nbformat == nbformat


def test_empty_markdown_cell():
    cell = new_markdown_cell()
    assert cell.cell_type == "markdown"
    assert cell.source == ""


def test_markdown_cell():
    cell = new_markdown_cell("* Søme markdown")
    assert cell.source == "* Søme markdown"


def test_empty_raw_cell():
    cell = new_raw_cell()
    assert cell.cell_type == "raw"
    assert cell.source == ""


def test_raw_cell():
    cell = new_raw_cell("hi")
    assert cell.source == "hi"


def test_empty_code_cell():
    cell = new_code_cell("hi")
    assert cell.cell_type == "code"
    assert cell.source == "hi"


def test_empty_display_data():
    output = new_output("display_data")
    assert output.output_type == "display_data"


def test_empty_stream():
    output = new_output("stream")
    assert output.output_type == "stream"
    assert output.name == "stdout"
    assert output.text == ""


def test_empty_execute_result():
    output = new_output("execute_result", execution_count=1)
    assert output.output_type == "execute_result"


mimebundle = {
    "text/plain": "some text",
    "application/json": {"key": "value"},
    "image/svg+xml": "ABCDEF",
    "application/octet-stream": "ABC-123",
    "application/vnd.foo+bar": "Some other stuff",
}


def test_display_data():
    output = new_output("display_data", mimebundle)
    for key, expected in mimebundle.items():
        assert output.data[key] == expected


def test_execute_result():
    output = new_output("execute_result", mimebundle, execution_count=10)
    assert output.execution_count == 10
    for key, expected in mimebundle.items():
        assert output.data[key] == expected


def test_error():
    o = new_output(
        output_type="error",
        ename="NameError",
        evalue="Name not found",
        traceback=["frame 0", "frame 1", "frame 2"],
    )
    assert o.output_type == "error"
    assert o.ename == "NameError"
    assert o.evalue == "Name not found"
    assert o.traceback == ["frame 0", "frame 1", "frame 2"]


def test_code_cell_with_outputs():
    cell = new_code_cell(
        execution_count=10,
        outputs=[
            new_output("display_data", mimebundle),
            new_output("stream", text="hello"),
            new_output("execute_result", mimebundle, execution_count=10),
        ],
    )
    assert cell.execution_count == 10
    assert len(cell.outputs) == 3
    er = cell.outputs[-1]
    assert er.execution_count == 10
    assert er["output_type"] == "execute_result"


def test_stream():
    output = new_output("stream", name="stderr", text="hello there")
    assert output.name == "stderr"
    assert output.text == "hello there"
